#include "IZ_list.h"

/**
 * Initializes a list.
 * @param list - The list to initialize.
 */
void IZ_ListInitialize(IZ_List* list) {
	list->root = NULL;
	list->length = 0;
	list->iterator = NULL;
}

/**
 * Performs cleanup on a list.
 * @param list - The list to clean up.
 */
void IZ_ListTeardown(IZ_List* list) {
	while (list->length > 0) {
		IZ_ListDeleteNode(list->root);
	}
	IZ_free(list);
}

/**
 * Appends a node to the end of the list.
 * @param list - The list to append to.
 * @return Pointer to the newly created node.
 */
IZ_ListNode** IZ_ListAppendNode(IZ_List* list, void* node_value) {
	IZ_ListNode* new_node = IZ_malloc(sizeof(IZ_ListNode));
	new_node->value = node_value;
	new_node->next = NULL;
	new_node->list = list;
	new_node->list->length += 1;

	if (!(list->root)) {
		new_node->previous = NULL;
		new_node->list->root = new_node;
		return &new_node->list->root;
	}

	IZ_ListNode *last_node = new_node->list->root;
	while (last_node->next) {
		last_node = last_node->next;
	}

	new_node->previous = last_node;
	last_node->next = new_node;
	return &last_node->next;
}

/**
 * Deletes the first node in the list that matches the filter.
 * @param node - The node to delete.
 */
void IZ_ListDeleteNode(IZ_ListNode* node) {
	if (!node) {
		// should we raise warnings here?
		return;
	}

	if (node->previous) {
		node->previous->next = node->next;
	} else {
		if (node->next) {
			node->next->previous = NULL;
		}
		if (node->list) {
			node->list->root = node->next;
		}
	}

	if (node->list && node->list->length > 0) {
		node->list->length -= 1;
	}

	IZ_free(node);
}

/**
 * Finds the first node in the list that matches the filter.
 * @param list - The list to search.
 * @param filter - The filter to use to find the node.
 * @return Pointer to the node that matches the filter.
 */
IZ_ListNode** IZ_ListFindFirstNode(IZ_List* list, IZ_ListFindPredicate filter) {
	if (!(list && list->root)) {
		return NULL;
	}

	list->iterator = &list->root;
	u64 index = 0;
	do {
		if (!filter(list->iterator, index, list)) {
			list->iterator = &((*list->iterator)->next);
			index += 1;
			continue;
		}
		return list->iterator;
	} while (*list->iterator);

	return NULL;
}
